/**
 * The Bag contains all of the kobold's consumables and miscellaneous items.
 */
function Bag() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Converts an object of id: {amount, item} tuples into a new list of 
     * ItemAmount objects and returns that list.
     * @param {object} list The object of tuples to convert.
     * @returns {array} The list converted to an array of ItemAmounts.
     */
    function listToItemAmount(list) {
        var output = [];
        for (var key in list) {
            if (list.hasOwnProperty(key)) {
                var amount = list[key].amount;
                var item = constructItem(key);
                output.push(new ItemAmount(amount, item));
            }
        }
        return output;
    }

    /**
     * Fetches the id from the passed object.
     * If the passed object is already a number, returns it.
     * @param {object | int} objectOrId The object or the id.
     * @returns {int} The id.
     */
    function getItemId(objectOrId) {
        if (typeof (objectOrId === "object")) {
            return objectOrId.id;
        }
        return objectOrId;
    }

    /**
     * Converts the passed object or id into an object.
     * If it is already an object, it is returned.
     * If it is an item id, its constructor is fetched and called.
     * @param {object | ItemEnum} objectOrId The object or the id.
     * @returns {object} The (potentially constructed) object.
     */
    function convertToItem(objectOrId) {
        if (typeof objectOrId === "object") {
            return objectOrId;
        }
        return constructItem(objectOrId);
    }

    /**
     * Construct an item via its itemId and returns the item.
     * @param {ItemEnum} itemId The id of the item.
     * @returns {object} The constructed item.
     */
    function constructItem(itemId) {
        return ItemEnum.properties[itemId].create();
    }

    /**
     * Constructs an entry into the miscs of consumables objects.
     * @param {int} amount How much of the item the kobold has.
     * @param {object} item The properly-constructed item.
     * @returns {object} The created entry.
     */
    function constructEntry(amount, item) {
        return {"amount": amount, "item": item};
    }

    /**
     * Converts a JSON-loaded objects to an object containing properly-constructed items and returns them.
     * @param {object} list The JSON-loaded object.
     * @returns {object} The constructed object.
     */
    function loadList(list) {
        var output = {};
        for (var key in list) {
            if (list.hasOwnProperty(key)) {
                var amount = list[key].amount;
                var item = constructItem(key);
                output[key] = constructEntry(amount, item);
            }
        }
        return output;
    }


    // PUBLIC MEMBERS \\        this.public = "test";

    // Item-ID mapped objects containing objects with an "amount" and an "item" member.
    this.miscs = {};
    this.consumables = {};

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};


    /**
     * Adds the specified mount of the passed miscellaneous item to the kobold's bag.
     * @param {Miscellaneous | int} misc The miscellaneous item to add or the miscellaneous item's id.
     * @param {int} amount How many of the miscellaneous item to add. 1 assumed if not passed.
     */
    this.addMisc = function (misc, amount) {
        if(!amount) {
            amount = 1;
        }
        var id = getItemId(misc);
        if (that.miscs[id]) {
            that.miscs[id].amount += amount;
        } else {
            var item = convertToItem(misc);
            that.miscs[id] = constructEntry(amount, item);
        }
    };

    /**
     * Returns how many of the passed miscelleaneous item the kobold has.
     * @param {Miscellaneous | int} misc The miscellaneous item to check or the miscellaneous item's id.
     * @returns {int} How many of the miscellaneous item the kobold has.
     */
    this.hasMisc = function (misc) {
        var id = getItemId(misc);
        if (that.miscs[id]) {
            return that.miscs[id].amount;
        }
        return 0;
    };

    /**
     * Removes the specified amount of the passed miscellaneous item from the kobold's bag.
     * @param {Miscellaneous | int} misc The miscellaneous item to remove or the miscellaneous item's id.
     * @param {int} amount How many of the miscellaneous item to remove. 1 is assumed if not passed.
     */
    this.removeMisc = function (misc, amount) {
        if(!amount) {
            amount = 1;
        }
        var id = getItemId(misc);
        if (!that.miscs[id]) {
            return;
        }
        that.miscs[id].amount -= amount;
        if (that.miscs[id].amount <= 0) {
            delete that.miscs[id];
        }
    };


    /**
     * Removes all of the passed miscellaneous item from the kobold's bag.
     * @param {Miscellaneous | int} misc The miscellaneous item to remove or the miscellaneous item's id.
     */
    this.removeAllMisc = function (misc) {
        var id = getItemId(misc);
        delete miscs[id];
    };

    /**
     * Adds the specified amount of the passed consumable to the kobold's bag.
     * @param {Consumable | int} consumable The consumable to add or the consumable's item id.
     * @param {int} amount How many of the consumable to add. 1 assumed if not passed.
     */
    this.addConsumable = function (consumable, amount) {
        if(!amount) {
            amount = 1;
        }
        var id = getItemId(consumable);
        if (that.consumables[id]) {
            that.consumables[id].amount += amount;
        } else {
            var item = convertToItem(consumable);
            that.consumables[id] = constructEntry(amount, item);
        }
    };

    /**
     * Returns how many of the passed consumable the kobold has.
     * @param {Consumable | int} consumable The consumable to check or the consumable's item id.
     * @returns {int} How many of the consumable the kobold has.
     */
    this.hasConsumable = function (consumable) {
        var id = getItemId(consumable);
        if (that.consumables[id]) {
            return that.consumables[id].amount;
        }
        return 0;
    };

    /**
     * Removes the specified amount of passed consumable from the kobold's bag.
     * @param {Consumable | int} consumable The consumable to remove or the consumable's item id.
     * @param {int} amount How many of the consumable to remove. 1 is assumed if not passed.
     */
    this.removeConsumable = function (consumable, amount) {
        if(!amount) {
            amount = 1;
        }
        var id = getItemId(consumable);
        if (!that.consumables[id]) {
            return;
        }
        that.consumables[id].amount -= amount;
        if (that.consumables[id].amount <= 0) {
            delete that.consumables[id];
        }
    };

    /**
     * Removes all of the passed consumable from the kobold's bag.
     * @param {Consumable | int} consumable The consumable to remove or the consumable's item id.
     */
    this.removeAllConsumable = function (consumable) {
        var id = getItemId(consumable);
        delete consumables[id];
    };

    /**
     * Removes all miscellaneous items from the kobold's bag.
     */
    this.emptyMiscs = function () {
        that.miscs = {};
    };

    /**
     * Removes all consumables from the kobold's bag.
     */
    this.emptyConsumables = function () {
        that.consumables = {};
    };

    /**
     * Empties the bag, removing everything from it.
     */
    this.emptyBag = function () {
        that.emptyMiscs();
        that.emptyConsumables();
    };

    /**
     * Gets an array of ItemAmount objects representing all of the kobold's consumables. 
     * @returns {array} An array of ItemAmount objects representing all of the kobold's consumables. 
     */
    this.getConsumablesList = function () {
        return listToItemAmount(that.consumables);
    };

    /**
     * Gets an array of ItemAmount objects representing all of the kobold's miscellaneous items. 
     * @returns {array} An array of ItemAmount objects representing all of the kobold's miscellaneous items. 
     */
    this.getMiscItemsList = function () {
        return listToItemAmount(that.miscs);
    };

    /**
     * Loads this object from a type-stripped version thereof.
     * @param {Bag-stripped} bag A Bag object, stripped of private members and all functions.
     */
    this.load = function (bag) {
        that.miscs = loadList(bag.miscs);
        that.consumables = loadList(bag.consumables);
    };
}